#! /usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod

import attr

from .oscillator import Oscillator
from .util import ElementGetter, suffix_to_char

UNUSED_ATTRIBS = [
    "lpfMode",
    "voicePriority",
    "polyphonic",
    "transpose",
    "isArmedForRecording",
    "mode",
    "activeModFunction",
    "clippingAmount",
    "modFXType",
    "inputChannel",
    "currentFilterType",
    "modFXCurrentParam",
    "defaultVelocity",
]


@attr.s
class ClipInstance(object):
  # Start position and length, both in pulses (see PPQN).
  # E.g., start of 48 = 1 quarter note into the arranger.
  start = attr.ib()
  length = attr.ib()
  # Index into the list of clips.
  clip_idx = attr.ib()

  @staticmethod
  def from_hex(data):
    assert len(data) == 24

    return ClipInstance(
        start=int(data[0:8], 16),
        length=int(data[8:16], 16),
        clip_idx=int(data[16:24], 16),
    )


@attr.s
class Instrument(object):
  __metaclass__ = ABCMeta
  muted = attr.ib(type=bool)
  clip_instances = attr.ib(type=list)

  @abstractmethod
  def pretty_name(self):
    raise NotImplementedError

  @staticmethod
  def _parse_clip_instances(clip_instances_data):
    clip_instances = []

    # Remove 0x prefix.
    if len(clip_instances_data) >= 2 and clip_instances_data[:2] == "0x":
      clip_instances_data = clip_instances_data[2:]

    while clip_instances_data:
      clip_instances.append(ClipInstance.from_hex(clip_instances_data[0:24]))
      clip_instances_data = clip_instances_data[24:]

    return clip_instances


@attr.s
class AudioTrack(Instrument):
  name = attr.ib()

  def pretty_name(self):
    return self.name

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element, unused_attribs=UNUSED_ATTRIBS) as e:
      return AudioTrack(
          name=e.get_attrib("name"),
          muted=bool(e.get_attrib("isMutedInArrangement", int, 0)),
          clip_instances=Instrument._parse_clip_instances(
              e.get_attrib("clipInstances", str, "")))


@attr.s
class MidiChannel(Instrument):
  channel = attr.ib(type=int)
  suffix = attr.ib(type=str)

  def pretty_name(self):
    return "MIDI {}{}".format(self.channel + 1, self.suffix)

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element, unused_attribs=UNUSED_ATTRIBS) as e:
      return MidiChannel(
          channel=e.get_attrib("channel", int, 0),
          suffix=suffix_to_char(e.get_attrib("suffix", int, -1)),
          muted=bool(e.get_attrib("isMutedInArrangement", int, 0)),
          clip_instances=Instrument._parse_clip_instances(
              e.get_attrib("clipInstances", str, "")))


@attr.s
class CvChannel(Instrument):
  channel = attr.ib(type=int)
  suffix = attr.ib(type=str)

  def pretty_name(self):
    return "CV {}{}".format(self.channel + 1, self.suffix)

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element, unused_attribs=UNUSED_ATTRIBS) as e:
      return MidiChannel(
          channel=e.get_attrib("channel", int, 0),
          suffix=suffix_to_char(e.get_attrib("suffix", int, -1)),
          muted=bool(e.get_attrib("isMutedInArrangement", int, 0)),
          clip_instances=Instrument._parse_clip_instances(
              e.get_attrib("clipInstances", str, "")))


@attr.s
class Sound(Instrument):
  UNUSED_CHILDREN = [
      "lfo1", "lfo2", "unison", "compressor", "arpeggiator", "modKnobs"
  ]

  name = attr.ib()
  preset_slot = attr.ib()
  suffix = attr.ib()
  oscillators = attr.ib(factory=list)

  def pretty_name(self):
    if self.name:
      return self.name

    return "Synth {}{}".format(self.preset_slot, self.suffix)

  @classmethod
  def from_element(cls, element):
    with ElementGetter(
        element,
        unused_attribs=UNUSED_ATTRIBS,
        unused_children=Sound.UNUSED_CHILDREN) as e:
      sound = Sound(
          name=e.get_any_attrib(["presetName", "name"], str, ""),
          preset_slot=e.get_attrib("presetSlot", int, 0),
          suffix=suffix_to_char(e.get_attrib("presetSubSlot", int, -1)),
          muted=bool(e.get_attrib("isMutedInArrangement", int, 0)),
          clip_instances=Instrument._parse_clip_instances(
              e.get_attrib("clipInstances", str, "")),
          oscillators=e.get_any_children({
              "osc1": Oscillator.from_element,
              "osc2": Oscillator.from_element
          }))

    return sound


@attr.s
class Kit(Instrument):
  UNUSED_CHILDREN = ["selectedDrumIndex"]

  name = attr.ib()
  preset_slot = attr.ib()
  suffix = attr.ib()
  sound_sources = attr.ib(factory=list)

  def pretty_name(self):
    if self.name:
      return self.name

    return "Kit {}{}".format(self.preset_slot, self.suffix)

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element, unused_attribs=UNUSED_ATTRIBS) as e:
      kit = Kit(
          name=e.get_attrib("presetName", str, ""),
          preset_slot=e.get_attrib("presetSlot", int, 0),
          suffix=suffix_to_char(e.get_attrib("presetSubSlot", int, -1)),
          muted=bool(e.get_attrib("isMutedInArrangement", int, 0)),
          clip_instances=Instrument._parse_clip_instances(
              e.get_attrib("clipInstances", str, "")),
          sound_sources=e.get_child("soundSources", Kit._parse_sound_sources,
                                    []))

    return kit

  @staticmethod
  def _parse_sound_sources(element):
    sound_sources = []
    for child in element:
      if child.tag == "sound":
        sound_sources.append(Sound.from_element(child))
      else:
        print("Unsupported child tag in soundSources: {}".format(child.tag))
    return sound_sources



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from .clip import *
from .color import *
from .condition import *
from .instrument import *
from .note import *
from .oscillator import *
from .params import *
from .project import *
from .util import *



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod

import attr

from .util import ElementGetter, suffix_to_char

UNUSED_OSCILLATOR_ATTRIBS = []


@attr.s
class Oscillator(object):
  __metaclass__ = ABCMeta
  typ = attr.ib(type=str)

  @classmethod
  @abstractmethod
  def from_element(cls, element):
    with ElementGetter(element, nag=False) as e:
      typ = e.get_attrib("type", str, "")

      # TODO: Support FM (oscillator has no type).

      if typ == "sample":
        return SampleOscillator.from_element(element)
      else:
        print("Unsupported oscillator type: {}".format(typ))
        return Oscillator(typ=typ)


@attr.s
class SampleOscillator(Oscillator):
  file_path = attr.ib(default="")

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element, unused_attribs=UNUSED_OSCILLATOR_ATTRIBS) as e:
      sample_oscillator = SampleOscillator(
          typ=e.get_attrib("type", str),
          file_path=e.get_attrib("fileName", str, ""))

    # TODO: Parse zone.

    return sample_oscillator


#! /usr/bin/env python
# -*- coding: utf-8 -*-

import attr


@attr.s
class Color(object):
  r = attr.ib()
  g = attr.ib()
  b = attr.ib()


def section_to_color(section):
  if section == -1:
    return Color(255, 255, 255)

  section_colors = [
      Color(96, 192, 255),  # Light blue.
      Color(255, 64, 255),  # Light pink.
      Color(255, 192, 64),  # Light yellow.
      Color(96, 255, 192),  # Light ~seafoam~.
      Color(255, 0, 0),  # Red.
      Color(192, 255, 96),  # Greenish yellow.
      Color(0, 0, 255),  # blu.
      Color(255, 96, 0),  # Orange.
      Color(96, 96, 255),  # Blue again.
      Color(96, 255, 0),  # Green.
      Color(96, 255, 96),  # Green again.
      Color(128, 64, 255),  # Purple.
  ]
  return section_colors[section % 12]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
import itertools
import math
import random


def add_complex_conditions(notes):
  _add_stacked_note_conditions(notes)
  _add_dotted_note_conditions(notes)


class Condition(object):
  __metaclass__ = ABCMeta

  @abstractmethod
  def active(self, iteration):
    raise NotImplementedError

  @abstractmethod
  def static(self):
    raise NotImplementedError

  @classmethod
  def from_value(cls, x):
    assert 0x00 <= x <= 0xFF

    if x == 0x14:
      return NoCondition()
    if x & 0b10000000 or x < 0x14:
      return ProbabilityCondition.from_value(x)
    if x > 0x14 and x & 0b10000000 == 0:
      return IterationCondition.from_value(x)

    print("Unexpected condition value: {}".format(x))

    return NoCondition()


class NoCondition(Condition):
  """Condition for notes with *no* condition."""

  def __init__(self):
    return

  def active(self, iteration):
    return True

  def static(self):
    return True


class IterationCondition(Condition):
  """
  Condition for "X out of Y" conditions. E.g., trigger every 1 out of 2 iterations.
  """

  def __init__(self, numerator, denominator):
    self.numerator = numerator
    self.denominator = denominator

  def active(self, iteration):
    return (iteration % self.denominator) == (self.numerator - 1)

  def static(self):
    return False

  @classmethod
  def from_value(cls, x):
    assert x > 0x14
    # No high bit set.
    assert x & 0b10000000 == 0

    x = x - 0x14

    # Quadratic formula for (n * (n+1)) / 2 = x.
    denominator = math.floor((-1 + math.sqrt(1 + 8 * x)) / 2) + 1
    previous_base = ((denominator - 1) * (denominator)) // 2
    numerator = x - previous_base + 1

    return IterationCondition(numerator, denominator)


class ProbabilityCondition(Condition):
  """Condition for notes with probabilistic condition."""

  def __init__(self, probability, dotted):
    """Probability between 0 and 1."""
    self.probability = probability
    self.dotted = dotted

  def active(self, iteration):
    # E.g., if probability = 0.25 (25%), we want to return true 25% of the time.
    return random.random() <= self.probability

  def static(self):
    return False

  @classmethod
  def from_value(cls, x):
    assert x < 0x14 or x & 0b10000000

    dotted = bool(x & 0b10000000)
    # Remove high bit.
    x &= ~0b10000000

    probability = x / 20.0

    return ProbabilityCondition(probability, dotted)


class SharedProbabilityCondition(Condition):
  """
  Condition for:
    - Stacks of notes with the same probability, where all of the notes play or
      none at all.
    - Dotted notes, where a note will trigger based on another note.
  """

  def __init__(self, condition):
    self.condition = condition
    self.active_this_iteration = False
    self.last_iteration = -1

  # Must be called in order (notes sorted by start).
  def active(self, iteration):
    if iteration != self.last_iteration:
      self.last_iteration = iteration
      self.active_this_iteration = self.condition.active(iteration)

    return self.active_this_iteration

  def static(self):
    return False


class SingleNoteFromProbabilityStackCondition(Condition):
  """
  Condition for stacks of notes where probability adds up to 100%. One note from
  the stack will ever be sounded for each iteration.
  """

  def __init__(self, stack_size):
    self.stack_size = stack_size
    self.note_to_trigger_in_stack = -1
    self.current_note_in_stack = -1
    self.last_iteration = -1

  # Must be called in order (notes sorted by start, then y).
  def active(self, iteration):
    if iteration != self.last_iteration:
      self.note_to_trigger_in_stack = random.randint(0, self.stack_size - 1)
      self.current_note_in_stack = 0

    triggered = False
    if self.current_note_in_stack == self.note_to_trigger_in_stack:
      triggered = True

    self.current_note_in_stack = (self.current_note_in_stack + 1) % iteration

    return triggered

  def static(self):
    return False


class InverseCondition(Condition):
  """Condition which negates the given condition's result."""

  def __init__(self, condition):
    self.condition = condition

  def active(self, iteration):
    return not self.condition.active(iteration)

  def static(self):
    return self.condition.static()


# Handles notes at the same start time:
#  - If their probabilities add up to 100%, only play 1 of the notes.
#  - If the notes all have the same probability, always trigger together or not at all.
#    - If they add up to 100%, the logic above applies first.
def _add_stacked_note_conditions(notes):
  notes.sort(key=lambda note: (note.start, note.y))

  probability_notes = filter(
      lambda note: isinstance(note.condition, ProbabilityCondition), notes)

  # Group by start to iterate through stacks of notes.
  for _, note_stack in itertools.groupby(probability_notes,
                                         lambda note: note.start):
    note_stack = list(note_stack)
    if len(note_stack) <= 1:
      continue

    # If the notes' probabilities add up to 100%, only play 1 of the notes.
    summed_probabilities = sum(
        note.condition.probability for note in note_stack)

    if summed_probabilities == 1.0:
      print(
          "Warning: Not well-tested functionality. Stack of notes summed to 1.")
      condition = SingleNoteFromProbabilityStackCondition(len(note_stack))
      for note in note_stack:
        note.condition = condition
      continue

    # Do all the notes have the same probability?
    all_same_probability = all(
        note.condition.probability == note_stack[0].condition.probability
        for note in note_stack)

    if all_same_probability:
      print(
          "Warning: Not well-tested functionality. Stack of notes with same probability."
      )
      condition = SharedProbabilityCondition(note_stack[0].condition)
      for note in note_stack:
        note.condition = condition


# Handles dotted notes:
# - Trigger if the previous note with *same probability* was successfully triggered.
# - If probability is dotted and previous (1 - probability) note was *not* played, trigger.
def _add_dotted_note_conditions(notes):
  probability_notes = list(
      filter(lambda note: isinstance(note.condition, ProbabilityCondition),
             notes))
  probability_notes.sort(
      key=lambda note: (note.start, note.condition.probability))

  # Maps probability to lists of notes (sorted by start time).
  # Combines notes with complementary probabilities (e.g., 30% and 70%) together.
  probability_to_notes = {}
  for note in probability_notes:
    p = note.condition.probability
    if p > 0.5:
      p = 1.0 - p

    probability_to_notes.get(p, []).append(note)

  for p, notes_for_probability in probability_to_notes.items():
    q = 1.0 - p

    # TODO: Can dotted probability conditions be chained?
    # E.g.: 70% -> 70% -> 70% -> 70% -> 70%, and all only trigger if the first is triggered.
    # Latest non-dotted (base) note for the probability.
    latest_base_note_for_p = None
    # Latest non-dotted (base) note for the complementary probability:
    #   q = 1 - p.
    latest_base_note_for_q = None

    for note in notes_for_probability:
      note_p = note.condition.probability
      note_q = 1.0 - note_p

      if not note.dotted:
        if note_p == p:
          latest_base_note_for_p = note
        if note_q == q:
          latest_base_note_for_q = note
        continue

      # Otherwise, we have a dotted note.

      if (latest_base_note_for_p is not None and
          note_p == latest_base_note_for_p.condition.probability):
        # Note is dotted. Link it to the previous non-dotted note.
        print("Warning: Not well-tested functionality. Added dotted.")
        condition = SharedProbabilityCondition(latest_base_note_for_p.condition)
        latest_base_note_for_p.condition = condition
        note.condition = condition
      elif (latest_base_note_for_q is not None and
            note_q == latest_base_note_for_q.condition.probability):
        # Note is negative-dotted. Link it to the previous non-dotted complementary note.
        print(
            "Warning: Not well-tested functionality. Added dotted complementary."
        )
        condition = SharedProbabilityCondition(latest_base_note_for_q.condition)
        latest_base_note_for_q.condition = condition
        note.condition = InverseCondition(condition)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import attr
from .util import ElementGetter, unit_float_from_fixed_hex


@attr.s
class Params(object):
  volume = attr.ib(default=1.0)
  pan = attr.ib(default=0.0)

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element, nag=False) as e:
      return Params(
          volume=e.get_attrib("volume", unit_float_from_fixed_hex,
                              "0xE0000000"),
          pan=e.get_attrib("pan", unit_float_from_fixed_hex, "0x00000000") * 2.0
          - 1.0,
      )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import attr

from .condition import Condition, add_complex_conditions
from .note import Note, NoteRow
from .params import Params
from .util import ElementGetter, suffix_to_char


@attr.s
class Clip(object):
  length = attr.ib()
  colour_offset = attr.ib()
  section = attr.ib(default=-1)
  params = attr.ib(default=attr.Factory(Params))

  def has_audio(self):
    return False

  def has_midi(self):
    return False


@attr.s
class AudioClip(Clip):
  track_name = attr.ib()
  file_path = attr.ib()
  start_sample_pos = attr.ib()
  end_sample_pos = attr.ib()
  pitch_speed_independent = attr.ib(default=1)
  reversed = attr.ib(default=0)
  transpose = attr.ib(default=0.0)
  section = attr.ib(default=-1)
  params = attr.ib(default=attr.Factory(Params))

  # TODO: Add method to get path on disk (or file handle?).

  def has_audio(self):
    return True

  @classmethod
  def from_element(cls, element):
    with ElementGetter(
        element,
        unused_attribs=[
            "isPlaying", "priority", "overdubsShouldCloneAudioTrack",
            "isArmedForRecording", "isSoloing", "affectEntire", "selected",
            "attack", "beingEdited"
        ]) as e:

      audio_clip = AudioClip(
          track_name=e.get_attrib("trackName"),
          file_path=e.get_attrib("filePath"),
          start_sample_pos=e.get_attrib("startSamplePos", int),
          end_sample_pos=e.get_attrib("endSamplePos", int),
          length=e.get_attrib("length", int),
          colour_offset=e.get_attrib("colourOffset", int, 0),
          section=e.get_attrib("section", int, -1),
          pitch_speed_independent=bool(
              e.get_attrib("pitchSpeedIndependent", int, 1)),
          reversed=bool(e.get_attrib("reversed", int, 0)),
          transpose=e.get_attrib("transpose", float, 0.0),
      )

    for child in element:
      if child.tag == "params":
        audio_clip.params = Params.from_element(child)
      else:
        print("Unsupported child tag in audioClip: {}".format(child.tag))

    return audio_clip


@attr.s
class InstrumentClip(Clip):
  channel = attr.ib()
  preset_name = attr.ib()
  # Also known as: MIDI channel, instrument preset slot, CV channel.
  preset_slot = attr.ib()
  suffix = attr.ib()
  notes = attr.ib(factory=list)
  section = attr.ib(default=-1)
  params = attr.ib(default=attr.Factory(Params))

  def has_midi(self):
    return True

  @classmethod
  def from_element(cls, element):
    with ElementGetter(
        element,
        unused_attribs=[
            "yScroll",
            "yScrollKeyboard",
            "isArmedForRecording",
            "isSoloing",
            "inKeyMode",
            "isPlaying",
            "onKeyboardScreen",
            "affectEntire",
            "selected",
            "beingEdited",
            # TODO: Support.
            "midiBank",
            "midiPGM",
        ]) as e:
      instrument_clip = InstrumentClip(
          channel=e.get_attrib("midiChannel", int, 0),
          preset_name=e.get_attrib("instrumentPresetName", str, ""),
          preset_slot=e.get_any_attrib([
              "midiChannel",
              "instrumentPresetSlot",
              "cvChannel",
          ], int, -1),
          suffix=suffix_to_char(
              e.get_any_attrib([
                  "midiChannelSuffix",
                  "instrumentPresetSubSlot",
                  "cvChannelSuffix",
              ], int, -1)),
          length=e.get_attrib("length", int),
          colour_offset=e.get_attrib("colourOffset", int, 0),
          # Not present for arrange-only clips.
          section=e.get_attrib("section", int, -1),
      )

      for child in element:
        if child.tag == "noteRows":
          instrument_clip.notes = InstrumentClip._note_rows_to_notes(
              InstrumentClip._parse_note_rows(child))
        elif child.tag == "soundParams":
          # Not supported yet -- ignore.
          pass
        elif child.tag == "kitParams":
          # Not supported yet -- ignore.
          pass
        elif child.tag == "arpeggiator":
          # Not supported yet -- ignore.
          pass
        elif child.tag == "midiParams":
          # Not supported yet -- ignore.
          pass
        else:
          print("Unsupported child in instrumentClip: {}".format(child.tag))

      return instrument_clip

  @staticmethod
  def _parse_note_rows(element):
    note_rows = []
    for child in element:
      if child.tag == "noteRow":
        note_rows.append(NoteRow.from_element(child))
      else:
        print("Unsupported child in noteRows: {}".format(child.tag))
    return note_rows

  @staticmethod
  def _note_rows_to_notes(note_rows):
    notes = []

    for note_row in note_rows:
      y = note_row.y

      for note_row_note in note_row.notes:
        condition = Condition.from_value(note_row_note.condition_value)
        notes.append(
            Note(y, note_row_note.start, note_row_note.length,
                 note_row_note.velocity, condition, note_row.muted))

    add_complex_conditions(notes)

    return notes



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import attr

SAMPLE_RATE_HZ = 44100
PPQN = 48
SECONDS_PER_MINUTE = 60

PRINT_WARNINGS = True


def fixed_to_unit_float(x):
  # Convert unsigned value to signed.
  if x & 0x80000000:
    x = x - 2**32
  # See: http://blog.bjornroche.com/2009/12/int-float-int-its-jungle-out-there.html
  return float(x + 2**31) / 2**32


def pulses_to_seconds(pulses, tempo):
  # Tempo is in units of quarter notes / minute.
  return ((pulses / float(PPQN)) / tempo) * SECONDS_PER_MINUTE


def unit_float_from_fixed_hex(x):
  return fixed_to_unit_float(int(x, 16))


def suffix_to_char(suffix):
  if suffix == -1:
    return ""

  return chr(ord('A') + int(suffix))


@attr.s
class ElementGetter(object):
  element = attr.ib()
  nag = attr.ib(default=True)
  _attribs_gotten = attr.ib(default=attr.Factory(set))
  _unused_attribs = attr.ib(default=attr.Factory(set), converter=set)
  _children_gotten = attr.ib(default=attr.Factory(set))
  _unused_children = attr.ib(default=attr.Factory(set), converter=set)

  def __enter__(self):
    self._attribs_gotten = set()
    self._children_gotten = set()
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    for name in self.element.attrib:
      if (name not in self._unused_attribs and
          name not in self._attribs_gotten and self.nag):
        print("Warning: Unused attrib on {}: {}".format(self.element.tag, name))

    for child in self.element:
      if (child.tag not in self._unused_children and
          child.tag not in self._children_gotten and self.nag):
        print("Warning: Unused child on {}: {}".format(self.element.tag,
                                                       child.tag))

  def has_attrib(self, name):
    return name in self.element.attrib

  def get_any_attrib(self, names, converter=None, default=None):
    for name in names:
      if self.has_attrib(name):
        return self.get_attrib(name, converter, default)

    if default is None:
      raise KeyError("None of {} found in {}'s attrib: {}".format(
          names, self.element.tag, self.element.attrib))

    if converter is not None:
      return converter(default)

    return default

  def get_attrib(self, name, converter=None, default=None):
    v = None
    if name in self.element.attrib:
      v = self.element.attrib.get(name)
    elif default is not None:
      v = default
    else:
      raise KeyError("{} not found in attrib: {}".format(
          name, self.element.attrib))

    self._attribs_gotten.add(name)

    if converter is not None:
      return converter(v)
    return v

  def get_child(self, name, converter=None, default=None):
    c = None
    for child in self.element:
      if child.tag == name:
        c = child

    if c is None:
      if default is not None:
        c = default
      else:
        raise KeyError("{} child not found in {}".format(
            name, self.element.tag))

    self._children_gotten.add(name)

    if converter is not None:
      return converter(c)

    return c

  def get_any_children(self, name_to_converter, unknown_converter=None):
    children = []

    for child in self.element:
      if child.tag in name_to_converter:
        children.append(name_to_converter[child.tag](child))
        self._children_gotten.add(child.tag)
      elif unknown_converter is not None:
        children.append(unknown_converter(child))

    return children



import setuptools

with open('README.md') as fp:
    long_description = fp.read()

setuptools.setup(
    name='pydel',
    version='0.5.2',
    description='Parser for Synthstrom Audible Deluge song files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Dillon Cower',
    license='MIT',
    url='https://github.com/dcower/pydel',
    packages=setuptools.find_packages(),
    install_requires=[
        'attrs',
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
    ])


#! /usr/bin/env python
# -*- coding: utf-8 -*-

import attr

from .clip import AudioClip, InstrumentClip
from .instrument import AudioTrack, MidiChannel, CvChannel, Sound, Kit
from .util import ElementGetter, SAMPLE_RATE_HZ, PPQN, SECONDS_PER_MINUTE


@attr.s
class Project(object):
  tempo = attr.ib()
  instruments = attr.ib(factory=list)
  sections = attr.ib(factory=list)
  clips = attr.ib(factory=list)
  arrange_only_clips = attr.ib(factory=list)

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element) as e:
      deluge_project = Project(
          tempo=deluge_timer_to_tempo(
              e.get_attrib("timePerTimerTick", int),
              e.get_attrib("timerTickFraction", int)),
          instruments=e.get_child("instruments", Project._parse_instruments,
                                  []),
          sections=e.get_child("sections", Project._parse_sections, []),
          clips=e.get_child("sessionClips", Project._parse_clips, []),
          arrange_only_clips=e.get_child("arrangementOnlyTracks",
                                         Project._parse_clips, []))

    return deluge_project

  @staticmethod
  def _parse_instruments(element):
    with ElementGetter(element) as e:
      return e.get_any_children({
          "audioTrack": AudioTrack.from_element,
          "midiChannel": MidiChannel.from_element,
          "cvChannel": CvChannel.from_element,
          "sound": Sound.from_element,
          "kit": Kit.from_element,
      })

  @staticmethod
  def _parse_sections(element):
    # Ignore for now.
    return []

  @staticmethod
  def _parse_clips(element):
    with ElementGetter(element) as e:
      return e.get_any_children(
          {
              "audioClip": AudioClip.from_element,
              "instrumentClip": InstrumentClip.from_element,
          },
          # Append an empty clip to keep the instance indices in order.
          unknown_converter=lambda: None)


def deluge_timer_to_tempo(time_per_timer_tick, timer_tick_fraction):
  # Based on Downrush's calculation.
  # Get the tick fraction in floating point (0 = .0, 0xFFFFFFFF = .9999...)
  timer_tick_fraction = timer_tick_fraction / 0x100000000
  # How many samples are in each pulse.
  samples_per_pulse = time_per_timer_tick + timer_tick_fraction

  # Deluge uses 48 PPQN (pulses per quarter note), and a 44100Hz sampling rate.
  #
  # Example for 120 BPM:
  #  44100 samples     1 quarter note   60 seconds        1 pulse          120 quarter notes
  # --------------- * --------------- * ---------- * ------------------- = --------------------
  #    1 second          48 pulses       1 minute        459.375 samples       1 minute

  tempo = SAMPLE_RATE_HZ / PPQN * SECONDS_PER_MINUTE / samples_per_pulse

  return tempo



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import attr

from .condition import NoCondition
from .util import ElementGetter


def condition_expansion_needed(notes):
  """Determines whether expansion is needed; if not, don't expand."""

  return any(not note.condition.static() for note in notes)


def condition_expand_notes(notes, clip_length, instance_length):
  expanded_notes = []

  pos = 0
  note_idx = 0
  iteration = 0
  last_start = 0

  while pos < instance_length:
    note = notes[note_idx]

    note_idx = (note_idx + 1) % len(notes)
    pos += note.start - last_start

    if note.condition.active(iteration):
      expanded_notes.append(
          Note(note.y, pos, note.length, note.velocity, NoCondition(),
               note.muted))

    if note_idx == 0:
      last_start = 0
      iteration += 1
      # Round to the next iteration's start point.
      pos = clip_length * iteration
    else:
      last_start = note.start

  return expanded_notes


def notes_to_on_off_notes(notes):
  """
  Converts from a list of notes denoting start and length to a list of notes denoting
  on (velocity >= 0) and off (velocity = 0).
  """
  on_off_notes = []

  for note in notes:
    # Note on.
    on_off_notes.append(
        Note(note.y, note.start, 0, note.velocity, note.condition, note.muted))
    # Note off.
    on_off_notes.append(
        Note(note.y, note.start + note.length, 0, 0, note.condition,
             note.muted))

  return on_off_notes


@attr.s
class Note(object):
  y = attr.ib()
  start = attr.ib()
  length = attr.ib()
  velocity = attr.ib()
  condition = attr.ib()
  muted = attr.ib()


@attr.s
class NoteRowNote(object):
  start = attr.ib()
  length = attr.ib()
  velocity = attr.ib()
  condition_value = attr.ib()

  @classmethod
  def from_hex(cls, data):
    assert len(data) == 20

    return NoteRowNote(
        # Start position and length, both in pulses (see PPQN).
        # E.g., start of 48 = 1 quarter note into the arranger.
        start=int(data[0:8], 16),
        length=int(data[8:16], 16),
        velocity=int(data[16:18], 16),
        condition_value=int(data[18:20], 16),
    )


@attr.s
class NoteRow(object):
  y = attr.ib()
  muted = attr.ib()
  notes = attr.ib()

  @classmethod
  def from_element(cls, element):
    with ElementGetter(element, unused_attribs=["colourOffset"]) as e:
      return NoteRow(
          y=e.get_attrib("drumIndex" if e.has_attrib("drumIndex") else "y",
                         int),
          muted=bool(e.get_attrib("muted", int, 0)),
          notes=NoteRow._parse_note_data(e.get_attrib("noteData", str, "")),
      )

  @staticmethod
  def _parse_note_data(note_data):
    notes = []

    # Remove 0x prefix.
    if len(note_data) >= 2 and note_data[:2] == "0x":
      note_data = note_data[2:]

    while note_data:
      notes.append(NoteRowNote.from_hex(note_data[0:20]))
      note_data = note_data[20:]

    return notes
